#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.
"""This module provides a class for downloading sequences from genbank
called Genbank and an function that does the same thing called genbank.

The function can be used if the environmental variable **pydna_email** has
been set to a valid email address. The easiest way to do this permanantly is to edit the
`pydna.ini` file. See the documentation of :func:`pydna.open_config_folder`"""

from pydna.utils import memorize as _memorize
from pydna.genbankrecord import GenbankRecord as _GenbankRecord
from pydna.readers import read as _read

from Bio import Entrez as _Entrez
import re as _re
import os as _os
import logging as _logging

_module_logger = _logging.getLogger("pydna." + __name__)


# TODO http://httpbin.org/ use for testing?


class Genbank(object):
    """Class to facilitate download from genbank. It is easier and
    quicker to use the :func:`pydna.genbank.genbank` function directly.

    Parameters
    ----------
    users_email : string
        Has to be a valid email address. You should always tell
        Genbanks who you are, so that they can contact you.

    Examples
    --------

    >>> from pydna.genbank import Genbank
    >>> gb=Genbank("bjornjobb@gmail.com")
    >>> rec = gb.nucleotide("LP002422.1")   # <- entry from genbank
    >>> print(len(rec))
    1
    """

    def __init__(self, users_email: str, *args, tool="pydna", **kwargs):

        if not _re.match(
            r"[A-Z0-9._%+-]+@[A-Z0-9.-]+\.[A-Z]{2,4}", users_email, _re.IGNORECASE
        ):
            raise ValueError("email address {} is not valid.".format(users_email))

        _module_logger.info("#### Genbank ititiation ####")
        _module_logger.info("Genbank initiated with email: %s", users_email)
        _module_logger.info("Genbank initiated with tool : %s", tool)

        if users_email == "someone@example.com":
            raise ValueError(
                "you have to set your email address in order to download from Genbank"
            )
        self.email = users_email
        self.tool = tool

    def __repr__(self):
        """This method returns a short representation containing the email used to initiate."""
        return "GenbankConnection({})".format(self.email)

    @_memorize("pydna.genbank.Genbank.nucleotide")
    def nucleotide(self, item: str, seq_start=None, seq_stop=None, strand=1):

        """This method downloads a genbank nuclotide record from genbank. This method is
        cached by default. This can be controlled by editing the **pydna_cached_funcs** environment
        variable. The best way to do this permanently is to edit the edit the
        `pydna.ini` file. See the documentation of :func:`pydna.open_config_folder`

        Item is a string containing one genbank accession number
        for a nucleotide file. Genbank nucleotide accession numbers have this format:

        | A12345   = 1 letter  + 5 numerals
        | AB123456 = 2 letters + 6 numerals

        The accession number is sometimes followed by a point and version number

        | BK006936.2

        Item can also contain optional interval information in the following formats:

        | BK006936.2 REGION: complement(613900..615202)
        | NM_005546 REGION: 1..100
        | NM_005546 REGION: complement(1..100)
        | 21614549:1-100
        | 21614549:c100-1
        | 21614549 1-100
        | 21614549 c100-1

        It is useful to set an interval for large genbank records to limit the download time.
        The items above containing interval information and can be obtained directly by
        looking up an entry in Genbank and setting the `Change region shown` on the
        upper right side of the page. The `ACCESSION` line of the displayed Genbank
        file will have the formatting shown.

        Alternatively, seq_start and seq_stop can be set explicitly to the sequence intervals to be
        downloaded.

        If strand is 2. "c", "C", "crick", "Crick", "antisense","Antisense",
        "2", 2, "-" or "-1", the antisense (Crick) strand is returned, otherwise
        the sense (Watson) strand is returned.

        Result is returned as a :class:`pydna.genbankrecord.GenbankRecord` object.

        References
        ----------

        .. [#]   http://www.dsimb.inserm.fr/~fuchs/M2BI/AnalSeq/Annexes/Sequences/Accession_Numbers.htm
        .. [#]   http://www.ncbi.nlm.nih.gov/books/NBK25499/#chapter4.EFetch
        """
        matches = (
            (1, _re.search(r"(REGION:\s(?P<start>\d+)\.\.(?P<stop>\d+))", item)),
            (2, _re.search(r"(REGION: complement\((?P<start>\d+)\.\.(?P<stop>\d+)\))", item),),
            (1, _re.search(r"(:|\s)(?P<start>\d+)-(?P<stop>\d+)", item)),
            (2, _re.search(r"(:|\s)c(?P<start>\d+)-(?P<stop>\d+)", item)),
        )

        for strand_, match in matches:
            if match:
                seq_start = match.group("start")
                seq_stop = match.group("stop")
                item = item[: match.start()]
                strand = strand_
                break

        if strand not in [1, 2]:
            try:
                strand = {"c": 2,
                          "crick": 2,
                          "antisense": 2,
                          "2": 2,
                          "-": 2,
                          "-1": 2}[strand.lower()]
            except (KeyError, AttributeError):
                strand = 1

        _module_logger.info("#### Genbank download ####")
        _module_logger.info("item  %s", item)
        _module_logger.info("start %s", seq_start)
        _module_logger.info("stop  %s", seq_stop)

        _module_logger.info("strand  %s", str(strand))

        _Entrez.email = self.email
        _Entrez.tool = self.tool

        seq_start = int(seq_start) if seq_start else None
        seq_stop = int(seq_stop) if seq_stop else None

        _module_logger.info("Entrez.email  %s", self.email)
        text = _Entrez.efetch(
            db="nuccore",
            id=item,
            rettype="gbwithparts",
            seq_start=seq_start,
            seq_stop=seq_stop,
            strand=strand,
            retmode="text",
        ).read()

        _module_logger.info("text[:160]  %s", text[:160])

        return _GenbankRecord(
            _read(text),
            item=item,
            start=seq_start,
            stop=seq_stop,
            strand=strand)


def genbank(accession: str = "CS570233.1", *args, **kwargs):
    """
    Download a genbank nuclotide record.

    This function takes the same paramenters as the
    :func:pydna.genbank.Genbank.nucleotide method. The email address stored
    in the `pydna_email` environment variable is used. The easiest way set
    this permanantly is to edit the `pydna.ini` file.
    See the documentation of :func:`pydna.open_config_folder`

    if no accession is given, a `very short Genbank
    entry <https://www.ncbi.nlm.nih.gov/nuccore/CS570233.1>`_
    is used as an example (see below). This can be useful for testing the
    connection to Genbank.

    Please note that this result is also cached by default by settings in
    the pydna.ini file.
    See the documentation of :func:`pydna.open_config_folder`

    ::

        LOCUS       CS570233                  14 bp    DNA     linear   PAT 18-MAY-2007
        DEFINITION  Sequence 6 from Patent WO2007025016.
        ACCESSION   CS570233
        VERSION     CS570233.1
        KEYWORDS    .
        SOURCE      synthetic construct
          ORGANISM  synthetic construct
                    other sequences; artificial sequences.
        REFERENCE   1
          AUTHORS   Shaw,R.W. and Cottenoir,M.
          TITLE     Inhibition of metallo-beta-lactamase by double-stranded dna
          JOURNAL   Patent: WO 2007025016-A1 6 01-MAR-2007;
                    Texas Tech University System (US)
        FEATURES             Location/Qualifiers
             source          1..14
                             /organism="synthetic construct"
                             /mol_type="unassigned DNA"
                             /db_xref="taxon:32630"
                             /note="This is a 14bp aptamer inhibitor."
        ORIGIN
                1 atgttcctac atga
        //

    """
    email = _os.getenv("pydna_email")
    _module_logger.info("#### genbank function called ####")
    _module_logger.info("email      %s", email)
    _module_logger.info("accession  %s", email)
    gb = Genbank(email)
    return gb.nucleotide(accession, *args, **kwargs)


if __name__ == "__main__":
    cached = _os.getenv("pydna_cached_funcs", "")
    _os.environ["pydna_cached_funcs"] = ""
    import doctest

    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)
    _os.environ["pydna_cached_funcs"] = cached
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

"""A DNA sequence widget."""

import sys
import os
from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QMainWindow
from PyQt5.QtWidgets import QGridLayout
from PyQt5.QtWidgets import QPlainTextEdit
from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QDesktopWidget
from PyQt5.QtGui import QFont
from PyQt5.QtCore import QSize
from multiprocessing import Process


class SequenceWindow(QMainWindow):
    """docstring."""

    def __init__(self, record):
        QMainWindow.__init__(self)
        self.setMinimumSize(QSize(1000, 100))

        qtRectangle = self.frameGeometry()
        centerPoint = QDesktopWidget().availableGeometry().center()
        qtRectangle.moveCenter(centerPoint)
        self.move(qtRectangle.topLeft())

        self.setWindowTitle(record.name)
        centralWidget = QWidget()
        self.setCentralWidget(centralWidget)
        gridLayout = QGridLayout()
        centralWidget.setLayout(gridLayout)
        font = QFont("Consolas")
        font.setPointSize(16)
        centralWidget.setFont(font)
        textw = QPlainTextEdit()

        text = str(record.seq)
        html = ""
        for fl in record.features[0].location.parts:
            html = (f"{text[:fl.start]}<span style='background-color:#ffff00;'>"
                    f"{text[fl.start:fl.end]}</span>"
                    f"{text[fl.end:]}")

        textw.appendHtml(html)
        textw.setReadOnly(True)
        gridLayout.addWidget(textw, 0, 0)

def run_app(record):
    if os.fork() != 0:
        return
    app = QtWidgets.QApplication(sys.argv)
    app.mainWin = SequenceWindow(record)
    app.mainWin.show()
    app.exec_()


if __name__ == '__main__':

    t = """
        LOCUS       myDNA                     12 bp    DNA     linear       02-JAN-2023
        DEFINITION  .
        ACCESSION
        VERSION
        SOURCE      .
          ORGANISM  .
        COMMENT
        COMMENT     ApEinfo:methylated:0
        FEATURES             Location/Qualifiers
              misc_feature    5..8
                              /locus_tag="New Feature"
                              /label="New Feature"
                              /ApEinfo_label="New Feature"
                              /ApEinfo_fwdcolor="cyan"
                              /ApEinfo_revcolor="green"
                              /ApEinfo_graphicformat="arrow_data {{0 0.5 0 1 2 0 0 -1 0
                              -0.5} {} 0} width 5 offset 0"
        ORIGIN
                1 atccGATCgg tc
        //
        """
    from pydna.readers import read

    s = read(t)

    s.features

    p = Process(target=run_app, args=(s,))
    p.start()
    p.join()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

"""This module provide functions for melting temperature calculations."""


import math as _math
from Bio.SeqUtils import MeltingTemp as _mt
from Bio.SeqUtils import gc_fraction as _GC

import textwrap as _textwrap
from pydna._pretty import pretty_str as _pretty_str

# See the documentation for Bio.SeqUtils.MeltingTemp for more details
# The 10X Taq Buffer with (NH4)2SO4 is commercialized by companies like
# ThermoFisher, although we make it ourselves
# 10X Buffer Composition
# 750 mM Tris-HCl (pH 8.8 at 25°C),
# 200 mM (NH4)2SO4,
# 0.1% (v/v) Tween 20.


def tm_default(
    seq,
    check=True,
    strict=True,
    c_seq=None,
    shift=0,
    nn_table=_mt.DNA_NN4,  # DNA_NN4: values from SantaLucia & Hicks (2004)
    tmm_table=None,
    imm_table=None,
    de_table=None,
    dnac1=500 / 2,  # I assume 500 µM of each primer in the PCR mix
    dnac2=500 / 2,  # This is what MELTING and Primer3Plus do
    selfcomp=False,
    Na=40,
    K=0,
    Tris=75.0,  # We use the 10X Taq Buffer with (NH4)2SO4 (above)
    Mg=1.5,  # 1.5 mM Mg2+ is often seen in modern protocols
    dNTPs=0.8,  # I assume 200 µM of each dNTP
    saltcorr=7,  # Tm = 81.5 + 0.41(%GC) - 600/N + 16.6 x log[Na+]
    func=_mt.Tm_NN,  # Used by Primer3Plus to calculate the product Tm.
):
    return func(
        seq,
        check=check,
        strict=strict,
        c_seq=c_seq,
        shift=shift,
        nn_table=nn_table,
        tmm_table=tmm_table,
        imm_table=imm_table,
        de_table=de_table,
        dnac1=dnac1,
        dnac2=dnac2,
        selfcomp=selfcomp,
        Na=Na,
        K=K,
        Tris=Tris,
        Mg=Mg,
        dNTPs=dNTPs,
        saltcorr=saltcorr,
    )


def tm_dbd(
    seq,
    check=True,
    strict=True,
    c_seq=None,
    shift=0,
    nn_table=_mt.DNA_NN3,
    tmm_table=None,
    imm_table=None,
    de_table=None,
    dnac1=250,
    dnac2=250,
    selfcomp=False,
    Na=50,
    K=0,
    Tris=0,
    Mg=1.5,
    dNTPs=0.8,
    saltcorr=1,
    func=_mt.Tm_NN,
):
    return func(
        seq,
        check=check,
        strict=strict,
        c_seq=c_seq,
        shift=shift,
        nn_table=nn_table,
        tmm_table=tmm_table,
        imm_table=imm_table,
        de_table=de_table,
        dnac1=dnac1,
        dnac2=dnac2,
        selfcomp=selfcomp,
        Na=Na,
        K=K,
        Tris=Tris,
        Mg=Mg,
        dNTPs=dNTPs,
        saltcorr=saltcorr,
    )


def tm_product(seq: str, K=0.050):
    """Tm calculation for the amplicon.

    according to:

    Rychlik, Spencer, and Rhoads, 1990, Optimization of the anneal
    ing temperature for DNA amplification in vitro
    http://www.ncbi.nlm.nih.gov/pubmed/2243783
    """
    tmp = 81.5 + 0.41 * _GC(seq) * 100 + 16.6 * _math.log10(K) - 675/len(seq)
    return tmp


def ta_default(fp: str,
               rp: str,
               seq: str,
               tm=tm_default,
               tm_product=tm_product):
    """Ta calculation.

    according to:

    Rychlik, Spencer, and Rhoads, 1990, Optimization of the anneal
    ing temperature for DNA amplification in vitro
    http://www.ncbi.nlm.nih.gov/pubmed/2243783

    The formula described uses the length and GC content of the product and
    salt concentration (monovalent cations)
    """
    return 0.3 * min((tm(fp), tm(rp))) + 0.7 * tm_product(seq) - 14.9


def ta_dbd(fp, rp, seq, tm=tm_dbd, tm_product=None):
    return min((tm(fp), tm(rp))) + 3


def program(amplicon, tm=tm_default, ta=ta_default):
    r"""Returns a string containing a text representation of a suggested
    PCR program using Taq or similar polymerase.

    ::

     |95°C|95°C               |    |tmf:59.5
     |____|_____          72°C|72°C|tmr:59.7
     |3min|30s  \ 59.1°C _____|____|60s/kb
     |    |      \______/ 0:32|5min|GC 51%
     |    |       30s         |    |1051bp

    """

    taq_extension_rate = 45  # seconds/kB PCR product length (1min/kb)
    extension_time_taq = max(30, int(taq_extension_rate*len(amplicon)/1000))
    # seconds

    f = _textwrap.dedent(
        r"""
        |95°C|95°C               |    |tmf:{tmf:.1f}
        |____|_____          72°C|72°C|tmr:{tmr:.1f}
        |3min|30s  \ {ta:.1f}°C _____|____|{rate}s/kb
        |    |      \______/{0:2}:{1:0>2}|5min|GC {GC}%
        |    |       30s         |    |{size}bp
        """.format(
            rate=taq_extension_rate,
            size=len(amplicon.seq),
            ta=round(
                ta(
                    amplicon.forward_primer.footprint,
                    amplicon.reverse_primer.footprint,
                    str(amplicon.seq),
                ),
                1,
            ),
            tmf=tm(amplicon.forward_primer.footprint),
            tmr=tm(amplicon.reverse_primer.footprint),
            GC=int(amplicon.gc()*100),
            *map(int, divmod(extension_time_taq, 60)),
        )).strip()

    return _pretty_str(f)


taq_program = program


def dbd_program(amplicon, tm=tm_dbd, ta=ta_dbd):
    r"""Text representation of a suggested PCR program.

    Using a polymerase with a DNA binding domain such as Pfu-Sso7d.

    ::

     |98°C|98°C               |    |tmf:53.8
     |____|_____          72°C|72°C|tmr:54.8
     |30s |10s  \ 57.0°C _____|____|15s/kb
     |    |      \______/ 0:15|5min|GC 51%
     |    |       10s         |    |1051bp


     |98°C|98°C      |    |tmf:82.5
     |____|____      |    |tmr:84.4
     |30s |10s \ 72°C|72°C|15s/kb
     |    |     \____|____|GC 52%
     |    |      3:45|5min|15058bp

    """
    PfuSso7d_extension_rate = 15  # seconds/kB PCR product
    extension_time_PfuSso7d = max(10,
                                  int(PfuSso7d_extension_rate
                                      * len(amplicon)/1000))  # seconds

    # The program returned is eaither a two step or three step progrem
    # This depends on the tm and length of the primers in the
    # original instructions from finnzyme. These do not seem to be

    # Ta calculation for enzymes with dsDNA binding domains like phusion or Pfu-Sso7d
    # https://www.finnzymes.fi/tm_determination.html

    tmf = tm(amplicon.forward_primer.footprint)
    tmr = tm(amplicon.reverse_primer.footprint)

    if tmf >= 69.0 and tmr >= 69.0:

        f = _textwrap.dedent(r"""
                              |98°C|98°C      |    |tmf:{tmf:.1f}
                              |____|____      |    |tmr:{tmr:.1f}
                              |30s |10s \ 72°C|72°C|{rate}s/kb
                              |    |     \____|____|GC {GC_prod}%
                              |    |     {0:2}:{1:0>2}|5min|{size}bp
                              """.format(
                rate=PfuSso7d_extension_rate,
                tmf=tmf,
                tmr=tmr,
                GC_prod=int(amplicon.gc()*100),
                size=len(amplicon.seq),
                *map(int, divmod(extension_time_PfuSso7d, 60)))).strip()
    else:
        f = _textwrap.dedent(
            r"""
             |98°C|98°C               |    |tmf:{tmf:.1f}
             |____|_____          72°C|72°C|tmr:{tmr:.1f}
             |30s |10s  \ {ta:.1f}°C _____|____|{rate}s/kb
             |    |      \______/{0:2}:{1:0>2}|5min|GC {GC}%
             |    |       10s         |    |{size}bp
             """.format(
             rate=PfuSso7d_extension_rate,
             size=len(amplicon.seq),
             ta=round(
              ta(amplicon.forward_primer.footprint,
                 amplicon.reverse_primer.footprint,
                 amplicon.seq,),
                 1,
                ),
                tmf=tmf,
                tmr=tmr,
                GC=int(amplicon.gc()*100),
                *map(int, divmod(extension_time_PfuSso7d, 60)))).strip()

    return _pretty_str(f)


pfu_sso7d_program = dbd_program


def Q5(primer: str, *args, **kwargs):
    """For Q5 Ta they take the lower of the two Tms and add 1C
    (up to 72C). For Phusion they take the lower of the two
    and add 3C (up to 72C).
    """
    raise NotImplementedError


def tmbresluc(primer: str, *args, primerc=500.0, saltc=50, **kwargs):
    """Returns the tm for a primer using a formula adapted to polymerases
    with a DNA binding domain, such as the Phusion polymerase.

    Parameters
    ----------

    primer : string
        primer sequence 5'-3'

    primerc : float
       primer concentration in nM), set to 500.0 nm by default.

    saltc : float, optional
       Monovalent cation concentration in mM, set to 50.0 mM by default.

    thermodynamics : bool, optional
        prints details of the thermodynamic data to stdout. For
        debugging only.

    Returns
    -------
    tm : float
        the tm of the primer

    """

    from . import _thermodynamic_data

    saltc = float(saltc) / 1000
    pri = primerc / 1e9
    dS = -12.4
    dH = -3400

    STR = primer.lower()

    for i in range(len(STR) - 1):
        n1 = ord(STR[i])
        n2 = ord(STR[i + 1])
        dH += _thermodynamic_data.dHBr[n1 - 97][n2 - 97]
        dS += _thermodynamic_data.dSBr[n1 - 97][n2 - 97]

    tm = (
        dH / (1.9872 * _math.log(pri / 1600) + dS)
        + (16.6 * _math.log(saltc)) / _math.log(10)
    ) - 273.15

    return tm


if __name__ == "__main__":
    import os as _os

    cached = _os.getenv("pydna_cached_funcs", "")
    _os.environ["pydna_cached_funcs"] = ""
    import doctest

    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)
    _os.environ["pydna_cached_funcs"] = cached



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

"""Assembly of sequences by Gateway recombination.

Given a list of sequences (Dseqrecords), all sequences are analyzed for
presence of att(P|B|L|R)N where N is 1,2,3 or 4.

A graph is constructed where the att sites form a nodes and
sequences separating att sites form edges.

The NetworkX package is used to trace linear and circular paths through the
graph.
"""
from Bio.SeqFeature import ExactPosition as _ExactPosition
from Bio.SeqFeature import SimpleLocation as _SimpleLocation
from Bio.SeqFeature import CompoundLocation as _CompoundLocation
from pydna.utils import rc as _rc
from pydna.utils import memorize as _memorize
from pydna._pretty import pretty_str as _pretty_str
from pydna.contig import Contig as _Contig
from pydna.common_sub_strings import common_sub_strings
from pydna.dseqrecord import Dseqrecord as _Dseqrecord
import networkx as _nx
from copy import deepcopy as _deepcopy
import itertools as _itertools
import logging as _logging

_module_logger = _logging.getLogger("pydna." + __name__)

ambiguous_dna_regex = {
    "A": "T",
    "C": "G",
    "G": "C",
    "T": "A",
    "M": "[ACM]",
    "R": "[AGR]",
    "W": "[ATW]",
    "S": "[CGS]",
    "Y": "[CTY]",
    "K": "[GTK]",
    "V": "[ACGVMSR]",
    "H": "[ACTHMYW]",
    "D": "[AGTDRWK]",
    "B": "[CGTBSKY]",
    "X": "X",
    "N": "[ACGTBDHKMNRSVWY]" }

atts = """
attP1 AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT GTACAAA AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG CMASTWT AAAGYWG
attP2 AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT GTACAAG AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG CMASTWT AAAGYWG
attP3 AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT GTATAAT AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG CMASTWT AAAGYWG
attP4 AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT GTATAGA AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG CMASTWT AAAGYWG
attP5 AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT GTATACA AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG CMASTWT AAAGYWG

attB1 CMASTWT GTACAAA AAAGYWG AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG
attB2 CMASTWT GTACAAG AAAGYWG AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG
attB3 CMASTWT GTATAAT AAAGYWG AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG
attB4 CMASTWT GTATAGA AAAGYWG AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG
attB5 CMASTWT GTATACA AAAGYWG AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG

attR1 CMASTWT GTACAAA AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT AAAGYWG
attR2 CMASTWT GTACAAG AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT AAAGYWG
attR3 CMASTWT GTATAAT AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT AAAGYWG
attR4 CMASTWT GTATAGA AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT AAAGYWG
attR5 CMASTWT GTATACA AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT AAAGYWG

attL1 AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT GTACAAA AAAGYWG CMASTWT AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG
attL2 AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT GTACAAG AAAGYWG CMASTWT AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG
attL3 AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT GTATAAT AAAGYWG CMASTWT AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG
attL4 AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT GTATAGA AAAGYWG CMASTWT AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG
attL5 AAATAATGATTTTATTTTGACTGATAGTGACCTGTTCGTTGCAACAMATTGATRAGCAATKMTTTYTTATAATGCCMASTTT GTATACA AAAGYWG CMASTWT AAAGYWGAACGAGAAACGTAAAATGATATAAATATCAATATATTAAATTAGATTTTGCATAAAAAACAGACTACATAATRCTGTAAAACACAACATATSCAGTCAYWWTG
"""


retable = str.maketrans(ambiguous_dna_regex)

for line in (line for line in atts.splitlines() if line.strip()):
    name, *parts = line.split()
    for part in parts:
        part.translate(retable)

class _Memoize(type):
    @_memorize("pydna.gateway.Gateway")
    def __call__(cls, *args, **kwargs):
        return super().__call__(*args, **kwargs)


class Gateway(object, metaclass=_Memoize):
    """Assembly of linear DNA fragments into linear or circular constructs.

    The Assembly is meant to replace the Assembly method as it
    is easier to use. Accepts a list of Dseqrecords (source fragments) to
    initiate an Assembly object. Several methods are available for analysis
    of overlapping sequences, graph construction and assembly.

    Parameters
    ----------
    fragments : list
        a list of Dseqrecord objects.
    """

    def __init__(self, molecules=None):
        self.molecules =molecules















"""
Created on Sat Aug 21 15:41:42 2021

@author: bjorn


https://en.wikipedia.org/wiki/Cre-Lox_recombination

13bp	      8bp	   13bp
ATAACTTCGTATA-NNNTANNN-TATACGAAGTTAT


Name	    13 bp  	        8 bp  	    13 bp
            Recognition     Spacer      Recognition
            Region          Region      Region

Wild-Type	ATAACTTCGTATA	ATGTATGC	TATACGAAGTTAT
lox 511	    ATAACTTCGTATA	ATGTATaC	TATACGAAGTTAT
lox 5171	ATAACTTCGTATA	ATGTgTaC	TATACGAAGTTAT
lox 2272	ATAACTTCGTATA	AaGTATcC	TATACGAAGTTAT
M2	        ATAACTTCGTATA	AgaaAcca	TATACGAAGTTAT
M3	        ATAACTTCGTATA	taaTACCA	TATACGAAGTTAT
M7	        ATAACTTCGTATA	AgaTAGAA	TATACGAAGTTAT
M11	        ATAACTTCGTATA	cgaTAcca	TATACGAAGTTAT
lox 71	    TACCGTTCGTATA	NNNTANNN	TATACGAAGTTAT
lox 66	    ATAACTTCGTATA	NNNTANNN	TATACGAACGGTA

"""
































"""

https://blog.addgene.org/plasmids-101-cre-lox

https://en.wikipedia.org/wiki/Cre-Lox_recombination

13bp	      8bp	   13bp
ATAACTTCGTATA-NNNTANNN-TATACGAAGTTAT


Name	    13 bp  	        8 bp  	    13 bp
            Recognition     Spacer      Recognition
            Region          Region      Region

Wild-Type	ATAACTTCGTATA	ATGTATGC	TATACGAAGTTAT
lox511	    ATAACTTCGTATA	ATGTATaC	TATACGAAGTTAT
lox5171	    ATAACTTCGTATA	ATGTgTaC	TATACGAAGTTAT
lox2272	    ATAACTTCGTATA	AaGTATcC	TATACGAAGTTAT
M2	        ATAACTTCGTATA	AgaaAcca	TATACGAAGTTAT
M3	        ATAACTTCGTATA	taaTACCA	TATACGAAGTTAT
M7	        ATAACTTCGTATA	AgaTAGAA	TATACGAAGTTAT
M11	        ATAACTTCGTATA	cgaTAcca	TATACGAAGTTAT
lox71	    TACCGTTCGTATA	NNNTANNN	TATACGAAGTTAT
lox66	    ATAACTTCGTATA	NNNTANNN	TATACGAACGGTA

"""


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.
"""Provides a function for downloading online text files"""

import textwrap as _textwrap
import requests as _requests
import os as _os
from pydna._pretty import pretty_str as _pretty_str
from pydna.utils import memorize as _memorize
import logging as _logging

_module_logger = _logging.getLogger("pydna." + __name__)


@_memorize("pydna.download.download_text")
def download_text(url):
    _module_logger.info("#### DOWNLOAD TEXT ####")
    _module_logger.info("url = %s", url)
    req = _requests.get(url)
    _module_logger.info("url = %s", str(req))
    result = _textwrap.dedent(req.text).strip()
    result = result.replace("\r\n", "\n").replace("\r", "\n")
    _module_logger.info("result[:160] = %s", result[:160])
    return _pretty_str(result)


if __name__ == "__main__":
    import os as _os

    cached = _os.getenv("pydna_cached_funcs", "")
    _os.environ["pydna_cached_funcs"] = ""
    import doctest

    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)
    _os.environ["pydna_cached_funcs"] = cached



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# doctest: +NORMALIZE_WHITESPACE
# doctest: +SKIP
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

"""docstring."""

from PIL import Image as _Image
from PIL import ImageDraw as _ImageDraw
import numpy as _np
import math as _math
from scipy.interpolate import CubicSpline as _CubicSpline
from pydna.ladders import GeneRuler_1kb_plus as _mwstd


def interpolator(mwstd):
    """docstring."""
    interpolator = _CubicSpline([len(fr) for fr in mwstd[::-1]],
                                [fr.rf for fr in mwstd[::-1]],
                                bc_type="natural",
                                extrapolate=False)
    interpolator.mwstd = mwstd
    return interpolator


def gel(samples=None,
        gel_length=600,
        margin=50,
        interpolator=interpolator(mwstd=_mwstd)):
    """docstring."""
    max_intensity = 256
    lane_width = 50
    lanesep = 10
    start = 10
    samples = samples or [interpolator.mwstd]
    width = int(60 + (lane_width + lanesep) * len(samples))
    lanes = _np.zeros((len(samples), gel_length), dtype=int)
    image = _Image.new("RGB", (width, gel_length), "#ddd")
    draw = _ImageDraw.Draw(image)
    draw.rectangle((0, 0, (width, gel_length)), fill=(0, 0, 0))
    scale = (gel_length - margin) / interpolator(min(interpolator.x))

    for labelsource in samples[0]:
        peak_centre = (interpolator(len(labelsource))) * scale - 5 + start
        label = f"{len(labelsource):<5} -"
        draw.text((2, peak_centre), label, fill=(255, 255, 255))

    for lane_number, lane in enumerate(samples):
        for band in lane:
            log = _math.log(len(band), 10)
            height = (band.m() / (240 * log)) * 1e10
            peak_centre = interpolator(len(band)) * scale + start
            max_spread = 10
            if len(band) <= 50:
                peak_centre += 50
                max_spread *= 4
                max_intensity /= 10
            band_spread = max_spread / log
            for i in range(max_spread, 0, -1):
                y1 = peak_centre - i
                y2 = peak_centre + i
                intensity = (
                    height * _math.exp(
                        -float(((y1 - peak_centre) ** 2)) / (2 * (band_spread ** 2))
                    ) * max_intensity)
                for y in range(int(y1), int(y2)):
                    try:
                        lanes[lane_number][y] += intensity
                    except IndexError:
                        pass

    for i, lane in enumerate(lanes):
        max_intensity = _np.amax(lanes[i])
        if max_intensity > 256:
            lanes[i] = _np.multiply(lanes[i], 256)
            lanes[i] = _np.divide(lanes[i], max_intensity)

    for i, lane in enumerate(lanes):
        x1 = 50 + i * (lane_width + lanesep)
        x2 = x1 + lane_width
        for y, intensity in enumerate(lane):
            y1 = y
            y2 = y + 1
            draw.rectangle((x1, y1, x2, y2), fill=(intensity,
                                                   intensity,
                                                   intensity))

    return image

    #  im.rotate(90, expand=1)
    #  im_invert = ImageOps.invert(im)


if __name__ == "__main__":
    import os as _os

    cached = _os.getenv("pydna_cached_funcs", "")
    _os.environ["pydna_cached_funcs"] = ""
    import doctest

    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)
    _os.environ["pydna_cached_funcs"] = cached



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Classes for nicer Jupyter output.

The pretty_str class is similar to str but has a _repr_pretty_ method
for for nicer string output in the IPython shell and Jupyter notebook.
"""

from prettytable import PrettyTable as _Pt
from prettytable import MARKDOWN as _md
from copy import copy as _copy

class pretty_str(str):
    """Thanks to Min RK, UC Berkeley for this."""

    def _repr_pretty_(self, p, cycle):
        p.text(self)


class PrettyTable(_Pt):
    """docstring."""

    def lol(self):
        """docstring."""
        return [self._field_names] + self._rows

    def __repr__(self):
        """docstring."""
        return self.get_string()

    def _repr_markdown_(self):
        c = _copy(self)
        c.set_style(_md)
        return pretty_str(c.get_string())


if __name__ == "__main__":
    import os as _os

    cached = _os.getenv("pydna_cached_funcs", "")
    _os.environ["pydna_cached_funcs"] = ""
    import doctest

    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)
    _os.environ["pydna_cached_funcs"] = cached



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

from pydna.dseqrecord import Dseqrecord
import os as _os
import logging as _logging
from Bio.Blast import NCBIWWW
from Bio.Blast import NCBIXML

_module_logger = _logging.getLogger("pydna." + __name__)


email = _os.getenv("pydna_email")
tool = "pydna"

def genbank_accession(s: str):
    """docstring."""
    s = Dseqrecord(s)

    NCBIWWW.email = email
    NCBIWWW.tool =  tool

    result_handle = NCBIWWW.qblast("blastn",
                                   "nt",
                                   str(s.seq),
                                   hitlist_size=1,
                                   alignments=1,
                                   descriptions=1,
                                   expect=1e-8,
                                   megablast=True,
                                   service="megablast",
                                   ungapped_alignment=True)

    blast_records = NCBIXML.read(result_handle)
    best_alignment, *rest = blast_records.alignments
    best_hsp, *rest = best_alignment.hsps
    dbs = best_hsp.sbjct
    start, stop = sorted((best_hsp.sbjct_start, best_hsp.sbjct_end))
    result = Dseqrecord(dbs,
                        linear=True,
                        id=s.name,
                        name=s.name,
                        description=(f"{best_alignment.accession} "
                                     f"REGION: {start}..{stop}"))
    return result


if __name__ == "__main__":
    cached = _os.getenv("pydna_cached_funcs", "")
    _os.environ["pydna_cached_funcs"] = ""
    import doctest

    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)
    _os.environ["pydna_cached_funcs"] = cached
    pass


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pydna']

package_data = \
{'': ['*']}

install_requires = \
['appdirs>=1.4.4',
 'biopython>=1.80',
 'networkx>=2.8.8',
 'prettytable>=3.5.0',
 'pyfiglet>=0.8.post1',
 'pyperclip>=1.8.2']

extras_require = \
{'download': ['pyparsing>=2.4.7', 'requests>=2.26.0'],
 'express': ['cai2>=1.0.5'],
 'gel': ['scipy>=1.8.0', 'matplotlib>=3.4.3', 'pillow>=8.4.0'],
 'gui': ['pyqt5>=5.15.0']}

setup_kwargs = {
    'name': 'pydna',
    'version': '5.2.0',
    'description': 'Representing double stranded DNA and functions for simulating cloning and homologous recombination between DNA molecules.',
    'long_description': '# ![icon](https://raw.githubusercontent.com/bjornFJohansson/pydna/master/docs/pics/pydna.resized.png) pydna\n\n| [![Tests & Coverage](https://github.com/BjornFJohansson/pydna/workflows/Tests%20&%20Coverage/badge.svg)](https://github.com/BjornFJohansson/pydna/actions?query=workflow%3A%22Tests+%26+Coverage%22)                |[![codecov](https://codecov.io/gh/BjornFJohansson/pydna/branch/master/graph/badge.svg)](https://codecov.io/gh/BjornFJohansson/pydna/branch/master)    | [![PyPI version](https://badge.fury.io/py/pydna.svg)](https://badge.fury.io/py/pydna)                                                   |[![Anaconda-Server Badge](https://anaconda.org/bjornfjohansson/pydna/badges/version.svg)](https://anaconda.org/bjornfjohansson/pydna)   | [![Google group : pydna](https://img.shields.io/badge/Google%20Group-pydna-blue.svg)](https://groups.google.com/g/pydna)        |\n|---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------|------------------------------------------------------------------------------------------------------------------------------------------------------|-----------------------------------------------------------------------------------------------------------------------------------------|----------------------------------------------------------------------------------------------------------------------------------------|---------------------------------------------------------------------------------------------------------------------------------|\n| [![Documentation Status](https://readthedocs.org/projects/pydna/badge/?version=latest)](http://pydna.readthedocs.io/?badge=latest)                                                                                  |[![GitHub issues](https://img.shields.io/github/issues/BjornFJohansson/pydna.svg)](https://github.com/BjornFJohansson/pydna/issues)                   | [![Anaconda-Server Badge2](https://anaconda.org/bjornfjohansson/pydna/badges/license.svg)](https://anaconda.org/bjornfjohansson/pydna)  |[![GitHub stars](https://img.shields.io/github/stars/BjornFJohansson/pydna.svg)](https://github.com/BjornFJohansson/pydna/stargazers)   |                                                                                                                                 |\n\n\nPlanning genetic constructs with many parts and assembly steps, such as recombinant\nmetabolic pathways :petri_dish:, are often difficult to **properly** document as is evident from the poor\nstate of documentation in the scientific literature :radioactive:.\n\n\nThe pydna python package provide a human-readable formal descriptions of :dna: cloning and genetic assembly\nstrategies in Python :snake: which allow for simulation and verification.\n\n\nPydna can be used as [executable documentation](https://en.wikipedia.org/wiki/Literate_programming) for cloning.\n\n\nA cloning strategy expressed in pydna is **complete**, **unambiguous** and **stable**.\n\n\nPydna provides simulation of:\n\n- Restriction digestion\n- Ligation\n- PCR\n- Primer design\n- Gibson assembly\n- Golden gate assembly\n- Homologous recombination\n- Gel electrophoresis of DNA with generation of gel images\n\nVirtually any sub-cloning experiment can be described in pydna, and its execution yield\nthe sequences of intermediate and final DNA molecules.\n\nPydna has been designed with the goal of being understandable for biologists with only some basic understanding of Python.\n\nPydna can formalize planning and sharing of cloning strategies and is especially useful for complex or combinatorial\nDNA molecule constructions.\n\n\nTo get started, we have compiled some [simple examples](https://github.com/MetabolicEngineeringGroupCBMA/pydna-examples#pydna-examples).\nFor more elaborate use, look at some assembly strategies of D-xylose metabolic pathways [MetabolicEngineeringGroupCBMA/ypk-xylose-pathways](https://github.com/MetabolicEngineeringGroupCBMA/ypk-xylose-pathways#pereira-et-al-2016).\n\n\n\n\n\n![-----------------------------------------------------](https://raw.githubusercontent.com/andreasbm/readme/master/assets/lines/colored.png)\n\n\n\n## Usage\n\nMost pydna functionality is implemented as methods for the double stranded DNA sequence record\nclasses Dseq and Dseqrecord, which are subclasses of the [Biopython](http://biopython.org/wiki/Main_Page) [Seq](http://biopython.org/wiki/Seq) and [SeqRecord](http://biopython.org/wiki/SeqRecord) classes.\n\nThese classes make cut and paste cloning and PCR very simple:\n\n    >>> from pydna.dseq import Dseq\n    >>> seq = Dseq("GGATCCAAA","TTTGGATCC",ovhg=0)\n    >>> seq\n    Dseq(-9)\n    GGATCCAAA\n    CCTAGGTTT\n    >>> from Bio.Restriction import BamHI\n    >>> a,b = seq.cut(BamHI)\n    >>> a\n    Dseq(-5)\n    G\n    CCTAG\n    >>> b\n    Dseq(-8)\n    GATCCAAA\n        GTTT\n    >>> a+b\n    Dseq(-9)\n    GGATCCAAA\n    CCTAGGTTT\n    >>> b+a\n    Dseq(-13)\n    GATCCAAAG\n        GTTTCCTAG\n    >>> b+a+b\n    Dseq(-17)\n    GATCCAAAGGATCCAAA\n        GTTTCCTAGGTTT\n    >>> b+a+a\n    Traceback (most recent call last):\n      File "<stdin>", line 1, in <module>\n      File "/usr/local/lib/python2.7/dist-packages/pydna/dsdna.py", line 217, in __add__\n        raise TypeError("sticky ends not compatible!")\n    TypeError: sticky ends not compatible!\n    >>>\n\nAs the example above shows, pydna keeps track of sticky ends.\n\nNotably, homologous recombination and Gibson assembly between linear DNA fragments\ncan be easily simulated without any additional information besides the primary sequence of the fragments.\n\nGel electrophoresis of DNA fragments can be simulated using the included gel module\n\n\n    Jupyter QtConsole 4.7.7\n    Python 3.8.5 | packaged by conda-forge | (default, Aug 29 2020, 01:22:49)\n    Type \'copyright\', \'credits\' or \'license\' for more information\n    IPython 7.18.1 -- An enhanced Interactive Python. Type \'?\' for help.\n\n    In [1]: from pydna.gel import gel\n\n    In [2]: from pydna.ladders import PennStateLadder\n\n    In [3]: from pydna.dseqrecord import Dseqrecord\n\n    In [4]: gel([PennStateLadder,[Dseqrecord("A"*2000)]])\n    Out[4]:\n\n\n\n![](https://raw.githubusercontent.com/BjornFJohansson/pydna/master/scripts/pydna_gel.png)\n\n\nPydna can be very compact. The eleven lines of Python below simulates the construction of a recombinant plasmid.\nDNA sequences are downloaded from Genbank by accession numbers that are guaranteed to be stable over time.\n\n    from pydna.genbank import Genbank\n    gb = Genbank("myself@email.com") # Tell Genbank who you are!\n    gene = gb.nucleotide("X06997") # Kluyveromyces lactis LAC12 gene for lactose permease.\n    from pydna.parsers import parse_primers\n    primer_f,primer_r = parse_primers(\'\'\' >760_KlLAC12_rv (20-mer)\n                                          ttaaacagattctgcctctg\n\n                                          >759_KlLAC12_fw (19-mer)\n                                          aaatggcagatcattcgag \'\'\')\n    from pydna.amplify import pcr\n    pcr_prod = pcr(primer_f,primer_r, gene)\n    vector = gb.nucleotide("AJ001614") # pCAPs cloning vector\n    from Bio.Restriction import EcoRV\n    lin_vector = vector.linearize(EcoRV)\n    rec_vec =  ( lin_vector + pcr_prod ).looped()\n\nPydna can automate the simulation of [sub cloning](http://en.wikipedia.org/wiki/Subcloning) experiments using\npython. This is helpful to generate examples for teaching purposes.\n\nRead the documentation (below) or the [cookbook](https://github.com/BjornFJohansson/pydna/blob/master/docs/cookbook/cookbook.ipynb) with example files\nfor further information.\n\nPlease post a message in the [google group](https://groups.google.com/d/forum/pydna)\nfor pydna if you need help or have problems, questions or comments :sos:.\n\nFeedback & suggestions are very welcome!\n\n![-----------------------------------------------------](https://raw.githubusercontent.com/andreasbm/readme/master/assets/lines/colored.png)\n\n\n## Who is using pydna?\n\nTaylor, L. J., & Strebel, K. (2017).\nPyviko: an automated Python tool to design gene knockouts in complex viruses with overlapping genes.\nBMC Microbiology, 17(1), 12.\n[PubMed](https://www.ncbi.nlm.nih.gov/pubmed/28061810)\n\n\nWang, Y., Xue, H., Pourcel, C., Du, Y., & Gautheret, D. (2021).\n2-kupl: mapping-free variant detection from DNA-seq data of matched samples.\nIn Cold Spring Harbor Laboratory (p. 2021.01.17.427048). [DOI](https://doi.org/10.1101/2021.01.17.427048)\n[PubMed](https://www.ncbi.nlm.nih.gov/pmc/articles/PMC8180056)\n\n\n[An Automated Protein Synthesis Pipeline with Transcriptic and Snakemake](http://blog.booleanbiotech.com/transcriptic_protein_synthesis_pipeline.html)\n\n\nand other projects on [github](https://github.com/BjornFJohansson/pydna/network/dependents?package_id=UGFja2FnZS01MjQ2MjYzNQ%3D%3D)\n\n![-----------------------------------------------------](https://raw.githubusercontent.com/andreasbm/readme/master/assets/lines/colored.png)\n\nThere is an open access paper in BMC Bioinformatics describing pydna:\n\n[![abstr](https://raw.githubusercontent.com/bjornFJohansson/pydna/master/docs/pics/BMC_resized.png)](http://www.biomedcentral.com/1471-2105/16/142/abstract)\n\nPlease reference the above paper:\n\n\nPereira, F., Azevedo, F., Carvalho, Â., Ribeiro, G. F., Budde, M. W., & Johansson, B. (2015). Pydna: a simulation and documentation tool for DNA assembly strategies using python. BMC Bioinformatics, 16(142), 142.\n\n\nWhen using pydna.\n\n![-----------------------------------------------------](https://raw.githubusercontent.com/andreasbm/readme/master/assets/lines/colored.png)\n\n## Documentation\n\nDocumentation is built using [Sphinx](http://www.sphinx-doc.org/) from [docstrings](https://www.python.org/dev/peps/pep-0257/)\nin the code and displayed at readthedocs [![Documentation Status](https://readthedocs.org/projects/pydna/badge/?version=latest)](http://pydna.readthedocs.io/?badge=latest)\n\nThe [numpy](www.numpy.org) [docstring format](https://github.com/numpy/numpy/blob/release/doc/HOWTO_DOCUMENT.rst.txt) is used.\n\n![-----------------------------------------------------](https://raw.githubusercontent.com/andreasbm/readme/master/assets/lines/colored.png)\n\n## Installation using pip\n\nPip is included in recent Python versions and is the\nofficially [recommended](http://python-packaging-user-guide.readthedocs.org/en/latest) tool.\n\nPip installs the minimal installation requirements automatically, but not the optional requirements (see below).\n\n    pip install pydna\n\nor use the --pre switch to get the latest version of pydna.\n\n    pip install pydna --pre\n\nfor optional functionality do:\n\n    pip install pydna[gel,download,express,gui]\n\nRemove options inside the square brackets as required, but be sure not to leave spaces as pip will not recognize the options. See below under "Optional dependencies".\n\n### Windows:\n\nYou should be able to pip install pydna from the Windows terminal as biopython now can be installed with pip as well.\n\n    C:\\> pip install pydna\n\nBy default python and pip are not on the PATH. You can re-install Python and select this option during installation, or give the full path for pip. Try something like this, depending on where your copy of Python is installed:\n\n    C:\\Python37\\Scripts\\pip install pydna\n\n![-----------------------------------------------------](https://raw.githubusercontent.com/andreasbm/readme/master/assets/lines/colored.png)\n\n## Source Code\n\nPydna is developed on [Github](https://github.com/BjornFJohansson/pydna) :octocat:.\n\n![-----------------------------------------------------](https://raw.githubusercontent.com/andreasbm/readme/master/assets/lines/colored.png)\n\n## Minimal installation dependencies\n\nPydna versions before 1.0.0 were compatible with python 2.7 only.\nThe list below is the minimal requirements for installing pydna.\nBiopython has c-extensions, but the other modules are pure python.\n\n- [Python 3.8, 3.9, 3.10 or 3.11](http://www.python.org)\n- [appdirs >=1.4.4](https://pypi.python.org/pypi/appdirs)\n- [biopython >= 1.80](http://pypi.python.org/pypi/biopython)\n- [networkx >=2.8.8](http://pypi.python.org/pypi/networkx)\n- [prettytable >=3.5.0](https://pypi.python.org/pypi/PrettyTable)\n- [pyperclip >=1.8.2](https://pypi.python.org/pypi/PrettyTable)\n- [pyfiglet >=0.8.post1](https://pypi.python.org/pypi/PrettyTable)\n\n\n## Optional dependencies\n\nIf the modules listed below in the first column are installed, they will provide the functionality listed in the second column.\n\n| Dependency                                                  | Function in pydna                                      |\n|-------------------------------------------------------------|--------------------------------------------------------|\n| [scipy >=1.8.0](https://www.scipy.org)                      | gel simulation with pydna.gel                          |\n| [matplotlib >=3.4.3](http://matplotlib.org)                 | “                                                      |\n| [pillow >=8.4.0](https://github.com/python-pillow/Pillow)   | “                                                      |\n| [numpy](http://www.numpy.org)                               | "                                                      |\n| [pyparsing >=2.4.7](https://pypi.python.org/pypi/pyparsing) | fix corrupt Genbank files with pydna.genbankfixer      |\n| [requests >=2.26.0](https://pypi.org/project/requests)      | download sequences with pydna.download                 |\n| [cai2 >=1.0.5](https://pypi.python.org/pypi/cai2)           | codon adaptation index calculations in several modules |\n| [pyqt5 >=5.15.0](https://pypi.python.org/pypi/pyqt5)        | codon adaptation index calculations in several modules |\n\n\n## Requirements for running tests and analyzing code coverage\n\n- [pytest](https://pypi.org/project/pytest)\n- [pytest-cov](https://pypi.org/project/pytest-cov)\n- [pytest-doctestplus](https://pypi.org/project/pytest-doctestplus)\n- [coverage](https://pypi.org/project/coverage)\n- [nbval](https://pypi.org/project/nbval)\n- [requests-mock](https://pypi.org/project/requests-mock)\n\n![----](https://raw.githubusercontent.com/andreasbm/readme/master/assets/lines/colored.png)\n\n## Releases\n\nSee the [releases](https://github.com/BjornFJohansson/pydna/releases) for changes and releases.\n\n![----](https://raw.githubusercontent.com/andreasbm/readme/master/assets/lines/colored.png)\n\n## Automatic testing & Release process\n\nThere are three github actions associated with this package:\n\n- `pydna_test_and_coverage_workflow.yml`\n\nThe `pydna_test_and_coverage_workflow.yml` is triggered on all pushed non-tagged commits.\nThis workflow run tests, doctests and a series of Jupyter notebooks using pytest.\n\nThe two other workflows build a setuptools wheel and packages for different Python versions\non Linux, Windows and macOS.\n\nThese are triggered by publishing a github release manually from the github interface.\n\n![----](https://raw.githubusercontent.com/andreasbm/readme/master/assets/lines/colored.png)\n\n## Building a PyPI package with [Poetry](https://pypi.org/project/poetry)\n\n1. Commit changes to git\n2. Tag the commit according to the [Semantic Versioning](https://semver.org) format, for example "v2.0.1a3". Do not forget the "v" or poetry will not recognize the tag.\n\n        git tag v2.0.1a3\n\n3. Pydna uses the poetry [poetry-dynamic-versioning](https://pypi.org/project/poetry-dynamic-versioning) plugin.\n\n        poetry dynamic-versioning # This sets the version number in the source files\n\n4. Verify the version\n\n        poetry version\n\n5. Build package:\n\n        poetry build # run this command in the root directory where the pyproject.toml file is located\n\n6. Verify the filename of the files in the dist/ folder, they should match\n\n7. Publish to pypi\n\n        poetry publish\n\n![----](https://raw.githubusercontent.com/andreasbm/readme/master/assets/lines/colored.png)\n\n## History\n\nPydna was made public in 2012 on [Google code](https://code.google.com/archive/p/pydna).\n\n\n:microbe:\n\n\n:portugal:\n',
    'author': 'BjornFJohansson',
    'author_email': 'bjornjobb@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.
"""Miscellaneous functions."""

from Bio.Data.IUPACData import (ambiguous_dna_complement as
                                _ambiguous_dna_complement)
from Bio.Seq import _maketrans
from pydna._pretty import pretty_str as _pretty_str
from Bio.SeqUtils.CheckSum import seguid as _seguid
import shelve as _shelve
import os as _os
import re as _re
import logging as _logging
import base64 as _base64
import pickle as _pickle
import hashlib as _hashlib
import keyword as _keyword
import collections as _collections
import itertools as _itertools

import sys as _sys
import re
import textwrap
import itertools
import math
import random
import subprocess as _subprocess

from pydna.codon import weights as _weights
from pydna.codon import rare_codons as _rare_codons
from pydna.codon import n_end as _n_end
from Bio.SeqUtils import seq3 as _seq3
from pydna._pretty import PrettyTable as _PrettyTable

from Bio.SeqFeature import SimpleLocation as _sl
from Bio.SeqFeature import CompoundLocation as _cl

_module_logger = _logging.getLogger("pydna." + __name__)
_ambiguous_dna_complement.update({"U": "A"})
_complement_table = _maketrans(_ambiguous_dna_complement)


def shift_location(original_location, shift, lim):
    """docstring."""
    newparts = []
    strand = original_location.strand
    for part in original_location.parts:
        ns = (part.start + shift) % lim
        ne = (part.end + shift) % lim or lim
        oe = newparts[-1].end if newparts else None
        if len(part) == 0:
            newparts.append(_sl(ns, ns, strand))
            continue
        elif oe == ns:
            part = newparts.pop()
            part._end = ne
            ns = part.start
        if ns < ne:
            newparts.append(_sl(ns, ne, strand))
        else:
            parttuple = (_sl(ns, lim, strand),
                         _sl(0,  ne,  strand))
            newparts.extend(parttuple if strand == 1 else parttuple[::-1])
    try:
        newloc = _cl(newparts)
    except ValueError:
        newloc, *n = newparts
    assert len(newloc) == len(original_location)
    return newloc


def smallest_rotation(s):
    """Smallest rotation of a string.

    Algorithm described in Pierre Duval, Jean. 1983. Factorizing Words
    over an Ordered Alphabet. Journal of Algorithms & Computational Technology
    4 (4) (December 1): 363–381. and Algorithms on strings and sequences based
    on Lyndon words, David Eppstein 2011.
    https://gist.github.com/dvberkel/1950267

    Examples
    --------
    >>> from pydna.utils import smallest_rotation
    >>> smallest_rotation("taaa")
    'aaat'

    """
    prev, rep = None, 0
    ds = 2 * s
    lens = len(s)
    lends = len(ds)
    old = 0
    k = 0
    w = ""
    while k < lends:
        i, j = k, k + 1
        while j < lends and ds[i] <= ds[j]:
            i = (ds[i] == ds[j]) and i + 1 or k
            j += 1
        while k < i + 1:
            k += j - i
            prev = w
            w = ds[old:k]
            old = k
            if w == prev:
                rep += 1
            else:
                prev, rep = w, 1
            if len(w) * rep == lens:
                return w * rep


def cai(seq: str,
        organism: str = "sce",
        weights: dict = _weights):
    """docstring."""
    from cai2 import CAI as _CAI
    return round(_CAI(seq.upper(), weights=weights[organism]), 3)


def rarecodons(seq: str,
               organism="sce"):
    """docstring."""
    rare = _rare_codons[organism]
    s = seq.upper()
    slices = []
    for i in range(0, len(seq)//3):
        x, y = i*3, i*3+3
        trip = s[x:y]
        if trip in rare:
            slices.append(slice(x, y, 1))
    return slices


def express(seq: str, organism="sce"):
    """docstring.


    **NOT IMPLEMENTED YET**
    """
    x = _PrettyTable(["cds", "len", "cai", "gc", "sta", "stp",
                      "n-end"]+_rare_codons[organism]+["rare"])
    val = []

    val.append(f"{self._data.upper().decode('ASCII')[:3]}..."
               f"{self._data.upper().decode('ASCII')[-3:]}")
    val.append(len(self)/3)
    val.append(cai(organism))
    val.append(gc())
    val.append(startcodon())
    val.append(stopcodon())
    val.append(_n_end[organism].get(_seq3(self[3:6].translate())))
    s = self._data.upper().decode("ASCII")
    trps = [s[i*3:i*3+3] for i in range(0, len(s)//3)]
    tot = 0
    for cdn in _rare_codons[organism]:
        cnt = trps.count(cdn)
        tot += cnt
        val.append(cnt)
    val.append(round(tot/len(trps), 3))
    x.add_row(val)
    return x


def open_folder(pth):
    """docstring."""
    if _sys.platform == "win32":
        _subprocess.run(["start", pth], shell=True)
    elif _sys.platform == "darwin":
        _subprocess.run(["open", pth])
    else:
        try:
            _subprocess.run(["xdg-open", pth])
        except OSError:
            return "no cache to open."


def rc(sequence: str):
    """Reverse complement.

    accepts mixed DNA/RNA
    """
    return sequence.translate(_complement_table)[::-1]


def complement(sequence: str):
    """Complement.

    accepts mixed DNA/RNA
    """
    return sequence.translate(_complement_table)


def memorize(filename):
    """Cache functions and classes.

    see pydna.download and pydna.Assembly for use
    """

    def decorator(f):
        def wrappee(*args, **kwargs):
            _module_logger.info("#### memorizer ####")
            _module_logger.info("cache filename                   = %s", filename)
            _module_logger.info(
                "os.environ['pydna_cached_funcs'] = %s",
                _os.environ["pydna_cached_funcs"],
            )
            if filename not in _os.environ["pydna_cached_funcs"]:
                _module_logger.info(
                    "cache filename not among cached functions, made it new!"
                )
                return f(*args, **kwargs)
            key = _base64.urlsafe_b64encode(
                _hashlib.sha1(_pickle.dumps((args, kwargs))).digest()
            ).decode("ascii")
            _module_logger.info("key = %s", key)
            cache = _shelve.open(
                _os.path.join(
                    _os.environ["pydna_data_dir"], identifier_from_string(filename)
                ),
                writeback=False,
            )
            try:
                result = cache[key]
            except KeyError:
                _module_logger.info(
                    "no result for key %s in shelve %s",
                    key,
                    identifier_from_string(filename),
                )
                result = f(*args, **kwargs)
                _module_logger.info("made it new!")
                cache[key] = result
                _module_logger.info("saved result under key %s", key)
            else:
                _module_logger.info("found %s in cache", key)
            cache.close()
            return result

        return wrappee

    return decorator


def identifier_from_string(s: str) -> str:
    """Return a valid python identifier.

    based on the argument s or an empty string
    """
    s = s.strip()
    s = _re.sub(r"\s+", r"_", s)
    s.replace("-", "_")
    s = _re.sub("[^0-9a-zA-Z_]", "", s)
    if s and not s[0].isidentifier() or _keyword.iskeyword(s):
        s = "_{s}".format(s=s)
    assert s == "" or s.isidentifier()
    return s


def seguid(seq: str) -> _pretty_str:
    """SEGUID checksum for a string representing a biological sequence.

    This is the SEGUID checksum with the standard Base64
    encoding that can contain '+' and '/' as originally
    defined by Babnigg and Giometti:

    Babnigg, Giometti 2006. “A Database of Unique
    Protein Sequence Identifiers for Proteome Studies.” *Proteomics* 6
    (16) (August): 4514–4522.

    Examples
    --------
    >>> from pydna.utils import seguid
    >>> seguid("aaa")
    'YG7G6b2Kj/KtFOX63j8mRHHoIlE'
    """
    return _pretty_str(_seguid(seq.upper()))

def useguid(seq: str) -> _pretty_str:
    """Returns the url safe SEGUID checksum for the sequence.
    This is the SEGUID checksum with the '+' and '/' characters of standard
    Base64 encoding are respectively replaced by '-' and '_'.

    Examples
    --------
    >>> from pydna.utils import useguid
    >>> useguid("aaa")
    'YG7G6b2Kj_KtFOX63j8mRHHoIlE'
    """
    return seguid(seq).replace("+", "-").replace("/", "_")


def lseguid_blunt(seq: str) -> _pretty_str:
    """lSEGUID checksum.

    for a string representing a blunt double stranded DNA molecule.

    Examples
    --------
    >>> from pydna.utils import lseguid_blunt
    >>> lseguid_blunt("ttt")
    'YG7G6b2Kj_KtFOX63j8mRHHoIlE'
    >>> lseguid_blunt("aaa")
    'YG7G6b2Kj_KtFOX63j8mRHHoIlE'
    """
    return useguid(min(seq.upper(), str(rc(seq)).upper()))


def lseguid_sticky(watson: str, crick: str, overhang: int) -> _pretty_str:
    """Linear SEGUID (lSEGUID) checksum.

    Calculates the lSEGUID checksum for a double stranded DNA sequence
    described by two strings (watson and crick) representing the two
    complementary DNA strands and an integer describing the stagger
    between the two strands in the 5' end.

    The overhang is defined as the amount of 3' overhang in the 5'
    side of the molecule. A molecule with 5' overhang has a negative
    value.

        dsDNA    ovhg

          nnn...    2
        nnnnn...

          nnnn...    1
        nnnnn...

        nnnnn...    0
        nnnnn...

        nnnnn...   -1
          nnnn...

        nnnnn...   -2
          nnn...


    """
    watson = watson.upper()
    crick = crick.upper()
    lw = len(watson)
    lc = len(crick)
    if overhang == 0 and lw == lc:
        return lseguid_blunt(watson)
    else:
        w, c, o = min(((watson, crick, overhang),
                       (crick, watson, lw - lc + overhang)))

    return useguid(f"{o*chr(32)}{w}\n{-o*chr(32)}{c[::-1]}")


def cseguid(seq: str) -> _pretty_str:
    """Url safe cSEGUID for a string representing a circular double stranded
    DNA molecule.

    The cSEGUID is the SEGUID checksum calculated for the lexicographically
    minimal string rotation of a DNA sequence. Only defined for circular
    sequences.

    Examples
    --------
    >>> from pydna.utils import cseguid
    >>> cseguid("attt")
    'oopV-6158nHJqedi8lsshIfcqYA'
    >>> cseguid("ttta")
    'oopV-6158nHJqedi8lsshIfcqYA'
    """
    return useguid(min(smallest_rotation(seq.upper()),
                       smallest_rotation(str(rc(seq)).upper())))


def flatten(*args):
    """Flattens an iterable of iterables.

    Down to str, bytes, bytearray or any of the pydna or Biopython seq objects
    """
    output = []
    args = list(args)
    while args:
        top = args.pop()
        if (
                isinstance(top, _collections.abc.Iterable)
                and not isinstance(top, (str, bytes, bytearray))
                and not hasattr(top, "reverse_complement")
                ):
            args.extend(top)
        else:
            output.append(top)
    return output[::-1]


def seq31(seq):
    """Turn a three letter code protein sequence into one with one letter code.

    The single input argument 'seq' should be a protein sequence using single
    letter codes, as a python string.

    This function returns the amino acid sequence as a string using the one
    letter amino acid codes. Output follows the IUPAC standard (including
    ambiguous characters B for "Asx", J for "Xle" and X for "Xaa", and also U
    for "Sel" and O for "Pyl") plus "Ter" for a terminator given as an
    asterisk.

    Any unknown
    character (including possible gap characters), is changed into 'Xaa'.

    Examples
    --------
    >>> from Bio.SeqUtils import seq3
    >>> seq3("MAIVMGRWKGAR*")
    'MetAlaIleValMetGlyArgTrpLysGlyAlaArgTer'
    >>> from pydna.utils import seq31
    >>> seq31('MetAlaIleValMetGlyArgTrpLysGlyAlaArgTer')
    'M  A  I  V  M  G  R  W  K  G  A  R  *'
    """
    threecode = {
        "Ala": "A",
        "Asx": "B",
        "Cys": "C",
        "Asp": "D",
        "Glu": "E",
        "Phe": "F",
        "Gly": "G",
        "His": "H",
        "Ile": "I",
        "Lys": "K",
        "Leu": "L",
        "Met": "M",
        "Asn": "N",
        "Pro": "P",
        "Gln": "Q",
        "Arg": "R",
        "Ser": "S",
        "Thr": "T",
        "Val": "V",
        "Trp": "W",
        "Tyr": "Y",
        "Glx": "Z",
        "Xaa": "X",
        "Ter": "*",
        "Sel": "U",
        "Pyl": "O",
        "Xle": "J",
    }

    nr_of_codons = int(len(seq) / 3)
    sequence = [seq[i * 3 : i * 3 + 3].title() for i in range(nr_of_codons)]
    padding = " " * 2
    return padding.join([threecode.get(aa, "X") for aa in sequence])


def parse_text_table(rawtable, tabs=4):

    table = textwrap.dedent(rawtable.expandtabs(tabs)).strip()
    max_row_length = max([len(row.strip()) for row in table.splitlines()])
    rows = [row.ljust(max_row_length) for row in table.splitlines()]
    table = "\n".join(rows)
    empty_column_regex = r"(?:\n?\s{%s}\n)+" % len(rows)
    transposed_table = "\n".join(["".join(c) for c in zip(*rows)])
    cols = re.split(empty_column_regex, transposed_table)
    list_of_lists_cr = []

    for col in cols:
        columnlist = [
            "".join(c).strip() for c in zip(*[a for a in col.splitlines() if a])
        ]
        maxlen = max([len(c) for c in columnlist])
        columnlist = [c.ljust(maxlen) for c in columnlist]
        list_of_lists_cr.append(columnlist)

    list_of_lists_rc = [list(i) for i in zip(*list_of_lists_cr)]

    formatted = _pretty_str("\n".join(" ".join(cell) for cell in
                            list_of_lists_rc))

    columnsplit = _pretty_str("\n|||\n".join(
        [
            "\n".join(
                [
                    x.strip()
                    for x in ["".join(c) for c in zip(*col.strip("\n").splitlines())]
                ]
            )
            for col in cols
        ]
    ))

    rowsplit = "\n---\n".join(["\n".join(a).strip() for a in zip(*list_of_lists_cr)])
    rowsplit = _pretty_str("\n".join(row.strip() for row in rowsplit.splitlines()))

    return (formatted,
            columnsplit,
            rowsplit,
            list_of_lists_rc,
            list_of_lists_cr,)


def join_list_to_table(rawlist):

    if "|||\n" in rawlist:
        raw_columns = rawlist.split("|||\n")
        cols = [col.splitlines() for col in raw_columns]
    elif "---\n" in rawlist:
        rawrows = rawlist.split("---\n")
        rows = [row.splitlines() for row in rawrows]
        cols = list(itertools.zip_longest(*rows, fillvalue=""))
    else:
        return None

    number_of_rows = max([len(col) for col in cols])
    formatted_cols = []

    for col in cols:
        # print col
        rows = [row.strip() or "\"" for row in col]
        width = max([len(row) for row in rows])

        rows = [row.ljust(width) for row in rows]
        rows += [" " * width] * (number_of_rows - len(rows))
        formatted_cols.append(rows)

    rows = list(zip(*formatted_cols))

    combinedlist = []

    for row in rows:
        combinedlist.append(" ".join(row))

    new_text = "\n".join(combinedlist)

    return _pretty_str(new_text)


def expandtolist(content):
    """docstring."""
    resultlist = []
    for line in re.finditer(r"(?P<item>[^\(\)]*?)(?P<brack>\[.*?\])", content):
        text2rep = line.group("item")
        bracket = line.group("brack")
        padding = max(
            [len(str(x).strip()) for x in re.split(r"\.\.|,",
                                                   bracket.strip("[ ]"))]
        )
        inbracket = [item.strip("[ ]") for item in bracket.split(",")]
        expanded = []

        for item in inbracket:
            if re.match(r"(\d+\.\.\d+)|([a-z]+\.\."
                        r"[a-z]+)|([A-Z]+\.\.[A-Z]+)", item):
                low, high = item.split(
                    "..",
                )
                if low.isdigit() and high.isdigit():
                    r = [
                        "{:{}d}".format(x, padding)
                        for x in range(int(low), 1 + int(high))
                    ]
                if (low.islower() and high.islower()) or (
                    low.isupper() and high.isupper()
                ):
                    r = [chr(a) for a in range(ord(low), 1 + ord(high))]
                expanded.extend(r)
            else:
                expanded.append(item.strip())

        resultlist.append([text2rep + x for x in expanded])

    ml = max([len(x) for x in resultlist])

    norm = []
    for r in resultlist:
        mp = int(math.ceil(float(ml) / float(len(r))))
        norm.append(list(itertools.chain.from_iterable(list(zip(*(r,) * mp)))))

    rt = ""
    for a in range(ml):
        rt += "".join([b[a] for b in norm]) + "\n"
    return _pretty_str(rt)


def randomRNA(length, maxlength=None):
    """docstring."""
    if maxlength and maxlength > length:
        length = int(round(random.triangular(length, maxlength)))
    return "".join([random.choice("GAUC") for x in range(length)])


def randomDNA(length, maxlength=None):
    """docstring."""
    if maxlength and maxlength > length:
        length = int(round(random.triangular(length, maxlength)))
    return "".join([random.choice("GATC") for x in range(length)])


def randomORF(length, maxlength=None):
    """docstring."""
    length -= 2
    if maxlength and maxlength > length:
        length = int(round(random.triangular(length, maxlength - 2)))

    cdns = (
        "TTT",
        "TTC",
        "TTA",
        "TTG",
        "TCT",
        "TCC",
        "TCA",
        "TCG",
        "TAT",
        "TAC",
        "TGT",
        "TGC",
        "TGG",
        "CTT",
        "CTC",
        "CTA",
        "CTG",
        "CCT",
        "CCC",
        "CCA",
        "CCG",
        "CAT",
        "CAC",
        "CAA",
        "CAG",
        "CGT",
        "CGC",
        "CGA",
        "CGG",
        "ATT",
        "ATC",
        "ATA",
        "ATG",
        "ACT",
        "ACC",
        "ACA",
        "ACG",
        "AAT",
        "AAC",
        "AAA",
        "AAG",
        "AGT",
        "AGC",
        "AGA",
        "AGG",
        "GTT",
        "GTC",
        "GTA",
        "GTG",
        "GCT",
        "GCC",
        "GCA",
        "GCG",
        "GAT",
        "GAC",
        "GAA",
        "GAG",
        "GGT",
        "GGC",
        "GGA",
        "GGG",
    )

    starts = ("ATG",)
    stops = ("TAA", "TAG", "TGA")

    return (
        random.choice(starts)
        + "".join([random.choice(cdns) for x in range(length)])
        + random.choice(stops)
    )


def randomprot(length, maxlength=None):
    """docstring."""
    if maxlength and maxlength > length:
        length = int(round(random.triangular(length, maxlength)))
    return "".join([random.choice("ACDEFGHIKLMNPQRSTVWY")
                    for x in range(length)])


def eq(*args, **kwargs):
    """Compare two or more DNA sequences for equality.

    Compares two or more DNA sequences for equality i.e. if they
    represent the same double stranded DNA molecule.

    Parameters
    ----------
    args : iterable
        iterable containing sequences
        args can be strings, Biopython Seq or SeqRecord, Dseqrecord
        or dsDNA objects.
    circular : bool, optional
        Consider all molecules circular or linear
    linear : bool, optional
        Consider all molecules circular or linear

    Returns
    -------
    eq : bool
        Returns True or False

    Notes
    -----
    Compares two or more DNA sequences for equality i.e. if they
    represent the same DNA molecule.

    Two linear sequences are considiered equal if either:

    1. They have the same sequence (case insensitive)
    2. One sequence is the reverse complement of the other

    Two circular sequences are considered equal if they are circular
    permutations meaning that they have the same length and:

    1. One sequence can be found in the concatenation of the other sequence with itself.
    2. The reverse complement of one sequence can be found in the concatenation of the other sequence with itself.

    The topology for the comparison can be set using one of the keywords
    linear or circular to True or False.

    If circular or linear is not set, it will be deduced from the topology of
    each sequence for sequences that have a linear or circular attribute
    (like Dseq and Dseqrecord).

    Examples
    --------
    >>> from pydna.dseqrecord import Dseqrecord
    >>> from pydna.utils import eq
    >>> eq("aaa","AAA")
    True
    >>> eq("aaa","AAA","TTT")
    True
    >>> eq("aaa","AAA","TTT","tTt")
    True
    >>> eq("aaa","AAA","TTT","tTt", linear=True)
    True
    >>> eq("Taaa","aTaa", linear = True)
    False
    >>> eq("Taaa","aTaa", circular = True)
    True
    >>> a=Dseqrecord("Taaa")
    >>> b=Dseqrecord("aTaa")
    >>> eq(a,b)
    False
    >>> eq(a,b,circular=True)
    True
    >>> a=a.looped()
    >>> b=b.looped()
    >>> eq(a,b)
    True
    >>> eq(a,b,circular=False)
    False
    >>> eq(a,b,linear=True)
    False
    >>> eq(a,b,linear=False)
    True
    >>> eq("ggatcc","GGATCC")
    True
    >>> eq("ggatcca","GGATCCa")
    True
    >>> eq("ggatcca","tGGATCC")
    True
    """
    args = flatten(args)  # flatten

    topology = None

    if "linear" in kwargs:
        if kwargs["linear"] is True:
            topology = "linear"
        if kwargs["linear"] is False:
            topology = "circular"
    elif "circular" in kwargs:
        if kwargs["circular"] is True:
            topology = "circular"
        if kwargs["circular"] is False:
            topology = "linear"
    else:
        topology = set([arg.circular if
                        hasattr(arg, "circular") else None for arg in args])

        if len(topology) != 1:
            raise ValueError("sequences have different topologies")
        topology = topology.pop()
        if topology in (False, None):
            topology = "linear"
        elif topology is True:
            topology = "circular"

    args = [arg.seq if hasattr(arg, "seq") else arg for arg in args]
    args_string_list = [
        arg.watson.lower() if hasattr(arg, "watson") else str(arg).lower()
        for arg in args
    ]

    length = set((len(s) for s in args_string_list))

    if len(length) != 1:
        return False
    same = True

    if topology == "circular":
        # force circular comparison of all given sequences
        for s1, s2 in _itertools.combinations(args_string_list, 2):
            if not (s1 in s2 + s2 or rc(s1) in s2 + s2):
                same = False
    elif topology == "linear":
        # force linear comparison of all given sequences
        for s1, s2 in _itertools.combinations(args_string_list, 2):
            if not (s1 == s2 or s1 == rc(s2)):
                same = False
    return same

if __name__ == "__main__":
    import os as _os

    cached = _os.getenv("pydna_cached_funcs", "")
    _os.environ["pydna_cached_funcs"] = ""
    import doctest

    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)
    _os.environ["pydna_cached_funcs"] = cached



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.
"""This module provide most pydna functionality in the local namespace.

Example
-------

>>> from pydna.all import *
>>> Dseq("aaa")
Dseq(-3)
aaa
ttt
>>> Dseqrecord("aaa")
Dseqrecord(-3)
>>> from pydna.all import __all__
>>> __all__
['Anneal', 'pcr', 'Assembly', 'genbank', 'Genbank', 'download_text\
', 'Dseqrecord', 'Dseq', 'read', 'read_primer', 'parse', 'parse_primers\
', 'ape', 'primer_design', 'assembly_fragments', 'circular_assembly_fragments\
', 'eq', 'gbtext_clean', 'PrimerList']
>>>
"""


__all__ = [
    "Anneal",
    "pcr",
    "Assembly",
    "genbank",
    "Genbank",
    "download_text",
    "Dseqrecord",
    "Dseq",
    "read",
    "read_primer",
    "parse",
    "parse_primers",
    "ape",
    "primer_design",
    "assembly_fragments",
    "circular_assembly_fragments",
    "eq",
    "gbtext_clean",
    "PrimerList",
]


from pydna.amplify import Anneal
from pydna.amplify import pcr
from pydna.assembly import Assembly
from pydna.genbank import genbank
from pydna.genbank import Genbank
from pydna.download import download_text
from pydna.dseqrecord import Dseqrecord
from pydna.dseq import Dseq
from pydna.readers import read
from pydna.readers import read_primer
from pydna.parsers import parse
from pydna.parsers import parse_primers
from pydna.editor import ape
from pydna.design import primer_design
from pydna.design import assembly_fragments
from pydna.design import circular_assembly_fragments
from pydna.utils import eq
from pydna.genbankfixer import gbtext_clean
from pydna.myprimers import PrimerList


if __name__ == "__main__":
    import os as _os

    cached = _os.getenv("pydna_cached_funcs", "")
    _os.environ["pydna_cached_funcs"] = ""
    import doctest

    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)
    _os.environ["pydna_cached_funcs"] = cached



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.
"""docstring."""


class FakeSeq(object):
    """docstring."""

    def __init__(self,
                 length,
                 n=50e-15,  # 50 fmol = 0.05 pmol
                 rf=0):
        self._length = int(length)
        self.n = n
        self.rf = rf

    def m(self):
        """Mass of the DNA molecule in grams."""
        # M(Da) * n (mol) = g
        return self.M() * self.n

    def M(self):
        """M grams/mol."""
        return (308.9 * self._length + 79.0) * 2

    def __len__(self):
        """docstring."""
        return self._length

    def __lt__(self, other):
        """docstring."""
        return self._length < len(other)

    def __repr__(self):
        """docstring."""
        return f"FakeSeq({self._length:.1e})"

    def __str__(self):
        """docstring."""
        return self.__repr__()


if __name__ == "__main__":
    import os as _os

    cached = _os.getenv("pydna_cached_funcs", "")
    _os.environ["pydna_cached_funcs"] = ""
    import doctest

    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)
    _os.environ["pydna_cached_funcs"] = cached



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2013-2023 by Björn Johansson.  All rights reserved.
# This code is part of the Python-dna distribution and governed by its
# license.  Please see the LICENSE.txt file that should have been included
# as part of this package.

"""This module is based on the Py-rstr-max package that
was written by Romain Brixtel (rbrixtel_at_gmail_dot_com)
(https://brixtel.users.greyc.fr) and is available from
https://code.google.com/p/py-rstr-max
https://github.com/gip0/py-rstr-max
the original code was covered by an MIT licence."""

from array import array as _array
import itertools as _itertools
from operator import itemgetter as _itemgetter


def radixpass(a, b, r, s, n, k):
    c = _array("i", [0] * (k + 1))
    for i in range(n):
        c[r[a[i] + s]] += 1

    somme = 0
    for i in range(k + 1):
        freq, c[i] = c[i], somme
        somme += freq

    for i in range(n):
        b[c[r[a[i] + s]]] = a[i]
        c[r[a[i] + s]] += 1


def direct_kark_sort(s):
    alphabet = [None] + sorted(set(s))
    k = len(alphabet)
    n = len(s)
    t = dict((c, i) for i, c in enumerate(alphabet))
    SA = _array("i", [0] * (n + 3))
    kark_sort(_array("i", [t[c] for c in s] + [0] * 3), SA, n, k)
    return SA[:n]


def kark_sort(s, SA, n, K):
    n0 = (n + 2) // 3
    n1 = (n + 1) // 3
    n2 = n // 3
    n02 = n0 + n2

    SA12 = _array("i", [0] * (n02 + 3))
    SA0 = _array("i", [0] * n0)

    s12 = [i for i in range(n + (n0 - n1)) if i % 3]
    s12.extend([0] * 3)
    s12 = _array("i", s12)

    radixpass(s12, SA12, s, 2, n02, K)
    radixpass(SA12, s12, s, 1, n02, K)
    radixpass(s12, SA12, s, 0, n02, K)

    name = 0
    c0, c1, c2 = -1, -1, -1
    for i in range(n02):
        if s[SA12[i]] != c0 or s[SA12[i] + 1] != c1 or s[SA12[i] + 2] != c2:
            name += 1
            c0 = s[SA12[i]]
            c1 = s[SA12[i] + 1]
            c2 = s[SA12[i] + 2]
        if SA12[i] % 3 == 1:
            s12[SA12[i] // 3] = name
        else:
            s12[SA12[i] // 3 + n0] = name

    if name < n02:
        kark_sort(s12, SA12, n02, name + 1)
        for i in range(n02):
            s12[SA12[i]] = i + 1
    else:
        for i in range(n02):
            SA12[s12[i] - 1] = i

    s0 = _array("i", [SA12[i] * 3 for i in range(n02) if SA12[i] < n0])
    radixpass(s0, SA0, s, 0, n0, K)

    p = j = k = 0
    t = n0 - n1
    while k < n:
        i = SA12[t] * 3 + 1 if SA12[t] < n0 else (SA12[t] - n0) * 3 + 2
        j = SA0[p] if p < n0 else 0

        if SA12[t] < n0:
            test = (
                (s12[SA12[t] + n0] <= s12[j // 3]) if (s[i] == s[j]) else (s[i] < s[j])
            )
        elif s[i] == s[j]:
            test = (
                s12[SA12[t] - n0 + 1] <= s12[j // 3 + n0]
                if (s[i + 1] == s[j + 1])
                else s[i + 1] < s[j + 1]
            )
        else:
            test = s[i] < s[j]

        if test:
            SA[k] = i
            t += 1
            if t == n02:
                k += 1
                while p < n0:
                    SA[k] = SA0[p]
                    p += 1
                    k += 1

        else:
            SA[k] = j
            p += 1
            if p == n0:
                k += 1
                while t < n02:
                    SA[k] = (
                        (SA12[t] * 3) + 1 if SA12[t] < n0 else ((SA12[t] - n0) * 3) + 2
                    )
                    t += 1
                    k += 1
        k += 1


class Rstr_max:
    def __init__(self):
        self.array_str = []

    def add_str(self, str_unicode):
        self.array_str.append(str_unicode)

    def step1_sort_suffix(self):
        char_frontier = chr(2)

        self.global_suffix = char_frontier.join(self.array_str)

        nbChars = len(self.global_suffix)
        init = [-1] * nbChars
        self.idxString = _array("i", init)
        self.idxPos = _array("i", init)
        self.endAt = _array("i", init)

        k = idx = 0
        for mot in self.array_str:
            last = k + len(mot)
            for p in range(len(mot)):
                self.idxString[k] = idx
                self.idxPos[k] = p
                self.endAt[k] = last
                k += 1
            idx += 1
            k += 1

        self.res = direct_kark_sort(self.global_suffix)

    def step2_lcp(self):
        n = len(self.res)
        init = [0] * n
        rank = _array("i", init)
        LCP = _array("i", init)

        s = self.global_suffix
        suffix__array = self.res
        endAt = self.endAt

        for i in range(len(self.array_str), n):
            v = self.res[i]
            rank[v] = i

        l = 0
        for j in range(n):
            if l > 0:
                l -= 1
            i = rank[j]
            j2 = suffix__array[i - 1]
            if i:
                while l + j < endAt[j] and l + j2 < endAt[j2] and s[j + l] == s[j2 + l]:
                    l += 1
                LCP[i - 1] = l
            else:
                l = 0
        self.lcp = LCP

    def step3_rstr(self):
        prev_len = 0
        idx = 0
        results = {}
        len_lcp = len(self.lcp) - 1

        class Stack:
            pass

        stack = Stack()
        stack._top = 0
        stack.lst_max = []

        # if len(self.res) == 0 :
        #  return {}

        pos1 = self.res[0]
        for idx in range(len_lcp):
            current_len = self.lcp[idx]
            pos2 = self.res[idx + 1]
            end_ = max(pos1, pos2) + current_len
            n = prev_len - current_len
            if n < 0:
                # pushMany
                stack.lst_max.append([-n, idx, end_])
                stack._top += -n
            elif n > 0:
                self.removeMany(stack, results, n, idx)
            elif stack._top > 0 and end_ > stack.lst_max[-1][-1]:
                # setMax
                stack.lst_max[-1][-1] = end_

            prev_len = current_len
            pos1 = pos2

        if stack._top > 0:
            self.removeMany(stack, results, stack._top, idx + 1)

        return results

    def removeMany(self, stack, results, m, idxEnd):
        prevStart = -1
        while m > 0:
            n, idxStart, maxEnd = stack.lst_max.pop()
            if prevStart != idxStart:
                # idStr = self.idxString[maxEnd-1]
                # pos = self.idxPos[maxEnd-1]
                id_ = (maxEnd, idxEnd - idxStart + 1)
                if id_ not in results or results[id_][0] < stack._top:
                    results[id_] = (stack._top, idxStart)
                prevStart = idxStart
            m -= n
            stack._top -= n
        if m < 0:
            stack.lst_max.append([-m, idxStart, maxEnd - n - m])
            stack._top -= m

    def go(self):
        self.step1_sort_suffix()
        self.step2_lcp()
        r = self.step3_rstr()
        return r


def common_sub_strings(stringx: str, stringy: str, limit=25):
    """Finds all common substrings between stringx and stringy
    longer than limit. This function is case sensitive.
    The substrings may overlap.

    returns a list of tuples describing the substrings
    The list is sorted longest -> shortest.

    Parameters
    ----------
    stringx : str
    stringy : str
    limit : int, optional

    Returns
    -------
    list of tuple
        [(startx1,starty1,length1),(startx2,starty2,length2), ...]

        startx1 = startposition in x, where substring 1 starts
        starty1 = position in y where substring 1 starts
        length1 = lenght of substring


    Examples
    --------

    >>> from pydna.common_sub_strings import common_sub_strings
    >>> common_sub_strings("gatgatttcggtagtta", "gtcagtatgtctatctatcgcg", limit=3)
    [(1, 6, 3), (7, 17, 3), (10, 4, 3), (12, 3, 3)]

    ::

        Overlaps   Symbols
        (1, 6,  3)   ---
        (7, 17, 3)   +++
        (10, 4, 3)   ...
        (12, 3, 3)   ===


                    ===
        gatgatttcggtagtta           stringx
         ---   +++...

            ...
        gtcagtatgtctatctatcgcg      stringy
           ===---        +++

    """

    rstr = Rstr_max()
    rstr.add_str(stringx + "&" + stringy)
    r = rstr.go()
    match = {}  # _defaultdict(int)
    for (offset_end, nb), (l, start_plage) in r.items():
        startsx = []
        startsy = []
        if l < limit:
            continue
        for o in range(start_plage, start_plage + nb):
            offset = rstr.idxPos[rstr.res[o]]
            if offset > len(stringx):
                startsy.append(offset - len(stringx) - 1)
            else:
                startsx.append(offset)

        for a, b in _itertools.product(startsx, startsy):
            match[(a, b)] = max(match.get((a, b)) or 0, l)

    match = [(key[0], key[1], val) for key, val in list(match.items())]

    match.sort()

    match.sort(key=_itemgetter(2), reverse=True)

    return match


def terminal_overlap(stringx: str, stringy: str, limit=15):
    """Finds the the flanking common substrings between stringx and stringy
    longer than limit. This means that the results only contains substrings
    that starts or ends at the the ends of stringx and stringy.

    This function is case sensitive.

    returns a list of tuples describing the substrings
    The list is sorted longest -> shortest.

    Parameters
    ----------
    stringx : str
    stringy : str
    limit : int, optional

    Returns
    -------
    list of tuple
        [(startx1,starty1,length1),(startx2,starty2,length2), ...]

        startx1 = startposition in x, where substring 1 starts
        starty1 = position in y where substring 1 starts
        length1 = lenght of substring


    Examples
    --------

    >>> from pydna.common_sub_strings import terminal_overlap
    >>> terminal_overlap("agctatgtatcttgcatcgta", "gcatcgtagtctatttgcttac", limit=8)
    [(13, 0, 8)]

    ::

                        <-- 8 ->
           <---- 13 --->
           agctatgtatcttgcatcgta                    stringx
                        gcatcgtagtctatttgcttac      stringy
                        0

    """
    return [
        m
        for m in common_sub_strings(stringx, stringy, limit)
        if (m[0] == 0 and m[1] + m[2] == len(stringy))
        or (m[1] == 0 and m[0] + m[2] == len(stringx))
    ]


if __name__ == "__main__":
    import os as _os

    cached = _os.getenv("pydna_cached_funcs", "")
    _os.environ["pydna_cached_funcs"] = ""
    import doctest

    doctest.testmod(verbose=True, optionflags=doctest.ELLIPSIS)
    _os.environ["pydna_cached_funcs"] = cached
